import os, numpy

from PyQt5.QtWidgets import QWidget, QVBoxLayout

from oasys.widgets import gui as oasysgui
from oasys.util.oasys_util import get_sigma, get_average

class HistogramData(object):
    scan_value = 0.0
    histogram = None
    bins = None
    offset = 0.0
    xrange = None
    fwhm = 0.0
    sigma = 0.0
    peak_intensity = 0.0
    integral_intensity = 0.0
    centroid = 0.0

    def __init__(self, histogram=None, bins=None, offset=0.0, xrange=None, fwhm=0.0, sigma=0.0, peak_intensity=0.0, integral_intensity=0.0, scan_value=0.0, centroid=0.0):
        self.histogram = histogram
        self.bins = bins
        self.offset = offset
        self.xrange = xrange
        self.fwhm = fwhm
        self.sigma = sigma
        self.peak_intensity = peak_intensity
        self.integral_intensity = integral_intensity
        self.scan_value = scan_value
        self.centroid = centroid

class HistogramDataCollection(object):

    data = None

    def __init__(self, histo_data=None):
        super().__init__()

        if not histo_data is None:
            self.add_reference_data(histo_data)

    def add_reference_data(self, histo_data=HistogramData()):
        if self.data is None:
            self.data = numpy.array([[[histo_data.scan_value]*len(histo_data.bins)], [histo_data.bins], [histo_data.histogram]])
        else:
            self.data = self.data.flatten()
            self.data = numpy.insert(self.data, [0, int(len(self.data)/3), int(2*len(self.data)/3)], [[histo_data.scan_value]*len(histo_data.bins), histo_data.bins, histo_data.histogram])
            self.data = self.data.reshape(3, int(len(self.data)/3))

    def replace_reference_data(self, histo_data=HistogramData()):
        if self.data is None:
            self.data = numpy.array([[[histo_data.scan_value]*len(histo_data.bins)], [histo_data.bins], [histo_data.histogram]])
        else:
            self.data[0, 0] = [histo_data.scan_value]*len(histo_data.bins)
            self.data[1, 0] = histo_data.bins
            self.data[2, 0] = histo_data.histogram

    def add_histogram_data(self, histo_data=HistogramData()):
        if self.data is None:
            self.data = numpy.array([[[histo_data.scan_value]*len(histo_data.bins)], [histo_data.bins], [histo_data.histogram]])
        else:
            self.data = numpy.append(self.data, numpy.array([[[histo_data.scan_value]*len(histo_data.bins)], [histo_data.bins], [histo_data.histogram]]), axis=1)

    def get_scan_values(self):
        return self.data[0, :][:, 0]

    def get_positions(self):
        return self.data[1, :]

    def get_intensities(self):
        return self.data[2, :]

    def get_histogram_data_number(self):
        return self.data.shape()[1]

    def get_scan_value(self, index):
        return self.data[0, index][0]

    def get_position(self, index):
        return self.data[1, index]

    def get_intensity(self, index):
        return self.data[2, index]

class StatisticalDataCollection(object):

    data = None

    def __init__(self, histo_data=None):
        super().__init__()

        if not histo_data is None:
            self.add_reference_data(histo_data)

    def add_reference_data(self, histo_data=HistogramData()):
        if self.data is None:
            self.data = numpy.array([[histo_data.scan_value], [histo_data.fwhm], [histo_data.sigma], [histo_data.peak_intensity], [histo_data.integral_intensity], [histo_data.centroid]])
        else:
            self.data = self.data.flatten()
            self.data = numpy.insert(self.data,
                                     [0, int(len(self.data)/6), int(2*len(self.data)/6), int(3*len(self.data)/6), int(4*len(self.data)/6), int(5*len(self.data)/6)],
                                     [histo_data.scan_value, histo_data.fwhm, histo_data.sigma, histo_data.peak_intensity, histo_data.integral_intensity, histo_data.centroid])
            self.data = self.data.reshape(6, int(len(self.data)/6))

    def replace_reference_data(self, histo_data=HistogramData()):
        if self.data is None:
            self.data = numpy.array([[histo_data.scan_value], [histo_data.fwhm], [histo_data.sigma], [histo_data.peak_intensity], [histo_data.integral_intensity], [histo_data.centroid]])
        else:
            self.data[0, 0] = histo_data.scan_value
            self.data[1, 0] = histo_data.fwhm
            self.data[2, 0] = histo_data.sigma
            self.data[3, 0] = histo_data.peak_intensity
            self.data[4, 0] = histo_data.integral_intensity
            self.data[5, 0] = histo_data.centroid

    def add_statistical_data(self, histo_data=HistogramData()):
        if self.data is None:
            self.data = numpy.array([[histo_data.scan_value], [histo_data.fwhm], [histo_data.sigma], [histo_data.peak_intensity], [histo_data.integral_intensity], [histo_data.centroid]])
        else:
            self.data = numpy.append(self.data, numpy.array([[histo_data.scan_value], [histo_data.fwhm], [histo_data.sigma], [histo_data.peak_intensity], [histo_data.integral_intensity], [histo_data.centroid]]), axis=1)

    def get_scan_values(self):
        return self.data[0, :]

    def get_fwhms(self):
        return self.data[1, :]

    def get_sigmas(self):
        return self.data[2, :]

    def get_centroids(self):
        return self.data[5, :]

    def get_absolute_peak_intensities(self):
        try:
            return self.data[3, :]
        except:
            return -1

    def get_absolute_integral_intensities(self):
        try:
            return self.data[4, :]
        except:
            return -1
        
    def get_relative_peak_intensities(self):
        try:
            return self.data[3, :]/self.data[3, 0]
        except:
            return -1

    def get_relative_integral_intensities(self):
        try:
            return self.data[4, :]/self.data[4, 0]
        except:
            return -1

    def get_stats_data_number(self):
        return self.data.shape()[1]

    def get_scan_value(self, index):
        return self.data[0, index]

    def get_fwhm(self, index):
        return self.data[1, index]

    def get_sigma(self, index):
        return self.data[2, index]

    def get_centroid(self, index):
        return self.data[5, index]

    def get_absolute_peak_intensity(self, index):
        return self.data[3, index]

    def get_absolute_integral_intensity(self, index):
        return self.data[4, index]

    def get_relative_peak_intensity(self, index):
        return self.data[3, index]/self.data[3, 0]

    def get_relative_integral_intensity(self, index):
        return self.data[4, index]/self.data[4, 0]

class DoublePlotWidget(QWidget):

    def __init__(self, parent=None):
        super(QWidget, self).__init__(parent=parent)

        self.plot_canvas = oasysgui.plotWindow(roi=False, control=False, position=True, logScale=False, fit=True)
        self.plot_canvas.setFixedWidth(700)
        self.plot_canvas.setFixedHeight(520)

        self.plot_canvas.setDefaultPlotLines(True)
        self.plot_canvas.setDefaultPlotPoints(True)

        self.ax2 = self.plot_canvas._backend.ax.twinx()

        layout = QVBoxLayout()

        layout.addWidget(self.plot_canvas)

        self.setLayout(layout)

    def plotCurves(self, x, y1, y2, title, xlabel, ylabel1, ylabel2):
        self.plot_canvas._backend.ax.clear()
        self.ax2.clear()

        self.plot_canvas.addCurve(x, y1, replace=False, color="b", symbol=".", ylabel=ylabel1, linewidth=1.5)
        self.plot_canvas.setGraphXLabel(xlabel)
        self.plot_canvas.setGraphTitle(title)
        self.plot_canvas._backend.ax.set_ylabel(ylabel1, color="b")

        self.ax2.plot(x, y2, "r.-")
        self.ax2.set_ylabel(ylabel2, color="r")

import h5py

def write_histo_and_stats_file_hdf5(histo_data=HistogramDataCollection(),
                                    stats=StatisticalDataCollection(),
                                    suffix="",
                                    output_folder=""):
    file = h5py.File(os.path.join(output_folder, "variable_scan_histogram_and_statistics" + suffix + ".hdf5"), "w")

    if not histo_data is None:
        histos = file.create_group("variable_scan_histograms")

        for scan_value, positions, intensities in zip(histo_data.get_scan_values(), histo_data.get_positions(), histo_data.get_intensities()):
            histogram = histos.create_group("histogram_" + str(scan_value) + suffix)
            histogram.create_dataset("positions", data=positions)
            histogram.create_dataset("intensities", data=intensities)

    if not stats is None:
        statistics = file.create_group("statistics")

        statistics.create_dataset("scan_values", data=stats.get_scan_values())
        statistics.create_dataset("fhwm", data=stats.get_fwhms())
        statistics.create_dataset("sigma", data=stats.get_sigmas())
        statistics.create_dataset("centroid", data=stats.get_centroids())
        statistics.create_dataset("absolute_peak_intensity", data=stats.get_absolute_peak_intensities())
        statistics.create_dataset("relative_peak_intensity", data=stats.get_relative_peak_intensities())
        statistics.create_dataset("absolute_integral_intensity", data=stats.get_absolute_integral_intensities())
        statistics.create_dataset("relative_integral_intensity", data=stats.get_relative_integral_intensities())

    file.flush()
    file.close()

def write_histo_and_stats_file(histo_data=HistogramDataCollection(),
                               stats=StatisticalDataCollection(),
                               suffix="",
                               output_folder=""):
    if not histo_data is None:
        for scan_value, positions, intensities in zip(histo_data.get_scan_values(), histo_data.get_positions(), histo_data.get_intensities()):

            file = open(os.path.join(output_folder, "histogram_" + str(scan_value) + suffix + ".dat"), "w")

            for position, intensity in zip(positions, intensities):
                file.write(str(position) + "   " + str(intensity) + "\n")

            file.flush()
            file.close()

    if not stats is None:
        file_fwhm = open(os.path.join(output_folder, "fwhm" + suffix + ".dat"), "w")
        file_sigma = open(os.path.join(output_folder, "sigma" + suffix + ".dat"), "w")
        file_centroid = open(os.path.join(output_folder, "centroid" + suffix + ".dat"), "w")
        file_peak_intensity = open(os.path.join(output_folder, "intensity" + suffix + ".dat"), "w")

        file_fwhm.write("scan_value " + "   " + "fwhm" +  "\n")
        file_sigma.write("scan_value " + "   " + "sigma" +  "\n")
        file_centroid.write("scan_value " + "   " + "centroid" +  "\n")
        file_peak_intensity.write("scan_value " + "   " + "absolute_peak_intensity" + "   " + "relative_peak_intensity"
                                  + "   " + "absolute_integral_intensity" + "   " + "relative_integral_intensity" +  "\n")

        for scan_value, \
            fwhm, sigma, centroid, \
            absolute_peak_intensity, relative_peak_intensity, \
            absolute_integral_intensity, relative_integral_intensity  in zip(stats.get_scan_values(),
                                                                             stats.get_fwhms(),
                                                                             stats.get_sigmas(),
                                                                             stats.get_centroids(),
                                                                             stats.get_absolute_peak_intensities(),
                                                                             stats.get_relative_peak_intensities(),
                                                                             stats.get_absolute_integral_intensities(),
                                                                             stats.get_relative_integral_intensities()):
            file_fwhm.write(str(scan_value) + "   " + str(fwhm) + "\n")
            file_sigma.write(str(scan_value) + "   " + str(sigma) + "\n")
            file_centroid.write(str(scan_value) + "   " + str(centroid) + "\n")
            file_peak_intensity.write(str(scan_value) + "   " + str(absolute_peak_intensity) + "   " + str(relative_peak_intensity)
                                      + "   " + str(absolute_integral_intensity) + "   " + str(relative_integral_intensity) +  "\n")

        file_fwhm.flush()
        file_sigma.flush()
        file_centroid.flush()
        file_peak_intensity.flush()

        file_fwhm.close()
        file_sigma.close()
        file_centroid.close()
        file_peak_intensity.close()



if __name__=="__main__":
    pass



import pkgutil
import inspect
import logging

from orangecanvas.registry import discovery

from oasys.menus.menu import OMenu

log = logging.getLogger(__name__)


class MenuRegistry(object):

    def __init__(self, menus=None):
        self.registry = menus if menus is not None else []

    def menus(self):
        return self.registry

    def addMenu(self, menu):
        self.registry.append(menu)


def omenus_from_package(package):
    try:
        package = discovery.asmodule(package)
    except ImportError:
        return
    for path in package.__path__:
        for _, mod_name, ispkg in pkgutil.iter_modules([path]):
            if ispkg:
                continue

            name = package.__name__ + "." + mod_name
            try:
                menu_module = discovery.asmodule(name)
            except ImportError:
                log.error("Error importing '%s'", name, exc_info=True)
                return

            for name, menu_class in inspect.getmembers(menu_module):
                if inspect.isclass(menu_class) and \
                        issubclass(menu_class, OMenu) and not name == "OMenu":
                    try:
                        yield menu_class()
                    except Exception:
                        log.error("Error creating menu instance from '%s'",
                                  menu_class, exc_info=True)



import pickle
from contextlib import contextmanager

from PyQt5.QtCore import  QItemSelectionModel
from PyQt5.QtCore import Qt, QAbstractListModel, QModelIndex, QByteArray
from PyQt5.QtCore import pyqtSignal as Signal

from PyQt5.QtWidgets import (
    QWidget, QBoxLayout, QToolButton, QAbstractButton, QAction
)

from oasys.util.script.variable import (
    Variable, DiscreteVariable, ContinuousVariable, StringVariable
)
from orangewidget import gui


class _store(dict):
    pass


def _argsort(seq, cmp=None, key=None, reverse=False):
    if key is not None:
        items = sorted(enumerate(seq), key=lambda i, v: key(v))
    elif cmp is not None:
        items = sorted(enumerate(seq), cmp=lambda a, b: cmp(a[1], b[1]))
    else:
        items = sorted(enumerate(seq), key=seq.__getitem__)
    if reverse:
        items = reversed(items)
    return items


@contextmanager
def signal_blocking(obj):
    blocked = obj.signalsBlocked()
    obj.blockSignals(True)
    yield
    obj.blockSignals(blocked)


def _as_contiguous_range(start, stop, step):
    if step == -1:
        # Equivalent range with positive step
        start, stop = stop + 1, start + 1
    elif not (step == 1 or step is None):
        raise IndexError("Non-contiguous range.")
    return start, stop, step


class PyListModel(QAbstractListModel):
    """ A model for displaying python list like objects in Qt item view classes
    """
    MIME_TYPES = ["application/x-Orange-PyListModelData"]

    def __init__(self, iterable=None, parent=None,
                 flags=Qt.ItemIsSelectable | Qt.ItemIsEnabled,
                 list_item_role=Qt.DisplayRole,
                 supportedDropActions=Qt.MoveAction):
        super().__init__(parent)
        self._list = []
        self._other_data = []
        self._flags = flags
        self.list_item_role = list_item_role

        self._supportedDropActions = supportedDropActions
        if iterable is not None:
            self.extend(iterable)

    def _is_index_valid_for(self, index, list_like):
        if isinstance(index, QModelIndex) and index.isValid():
            row, column = index.row(), index.column()
            return 0 <= row < len(list_like) and not column
        elif isinstance(index, int):
            return -len(self) < index < len(list_like)
        else:
            return False

    def wrap(self, lst):
        """ Wrap the list with this model. All changes to the model
        are done in place on the passed list
        """
        self._list = lst
        self._other_data = [_store() for _ in lst]
        #self.reset()


    # noinspection PyMethodOverriding
    def index(self, row, column=0, parent=QModelIndex()):
        if self._is_index_valid_for(row, self) and column == 0:
            return QAbstractListModel.createIndex(self, row, column, parent)
        else:
            return QModelIndex()

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            return str(section)


    # noinspection PyMethodOverriding
    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self)

    def columnCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else 1

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        if role in [self.list_item_role, Qt.EditRole] \
                and self._is_index_valid_for(index, self):
            return self[row]
        elif self._is_index_valid_for(row, self._other_data):
            return self._other_data[row].get(role, None)

    def itemData(self, index):
        mapping = QAbstractListModel.itemData(self, index)
        if self._is_index_valid_for(index, self._other_data):
            items = list(self._other_data[index.row()].items())
        else:
            items = []
        for key, value in items:
            mapping[key] = value
        return mapping

    def parent(self, index=QModelIndex()):
        return QModelIndex()

    def setData(self, index, value, role=Qt.EditRole):
        if role == Qt.EditRole and self._is_index_valid_for(index, self):
            self[index.row()] = value  # Will emit proper dataChanged signal
            return True
        elif self._is_index_valid_for(index, self._other_data):
            self._other_data[index.row()][role] = value
            self.dataChanged.emit(index, index)
            return True
        else:
            return False

    def setItemData(self, index, data):
        data = dict(data)
        with signal_blocking(self):
            for role, value in data.items():
                if role == Qt.EditRole and \
                        self._is_index_valid_for(index, self):
                    self[index.row()] = value
                elif self._is_index_valid_for(index, self._other_data):
                    self._other_data[index.row()][role] = value

        self.dataChanged.emit(index, index)
        return True

    def flags(self, index):
        if self._is_index_valid_for(index, self._other_data):
            return self._other_data[index.row()].get("flags", self._flags)
        else:
            return self._flags | Qt.ItemIsDropEnabled


    # noinspection PyMethodOverriding
    def insertRows(self, row, count, parent=QModelIndex()):
        """ Insert ``count`` rows at ``row``, the list fill be filled
        with ``None``
        """
        if not parent.isValid():
            self[row:row] = [None] * count
            return True
        else:
            return False


    # noinspection PyMethodOverriding
    def removeRows(self, row, count, parent=QModelIndex()):
        """Remove ``count`` rows starting at ``row``
        """
        if not parent.isValid():
            del self[row:row + count]
            return True
        else:
            return False

    def extend(self, iterable):
        list_ = list(iterable)
        self.beginInsertRows(QModelIndex(),
                             len(self), len(self) + len(list_) - 1)
        self._list.extend(list_)
        self._other_data.extend([_store() for _ in list_])
        self.endInsertRows()

    def append(self, item):
        self.extend([item])

    def insert(self, i, val):
        self.beginInsertRows(QModelIndex(), i, i)
        self._list.insert(i, val)
        self._other_data.insert(i, _store())
        self.endInsertRows()

    def remove(self, val):
        i = self._list.index(val)
        self.__delitem__(i)

    def pop(self, i):
        item = self._list[i]
        self.__delitem__(i)
        return item

    def __len__(self):
        return len(self._list)

    def __iter__(self):
        return iter(self._list)

    def __getitem__(self, i):
        return self._list[i]

    def __add__(self, iterable):
        new_list = PyListModel(list(self._list),
                               self.parent(),
                               flags=self._flags,
                               list_item_role=self.list_item_role,
                               supportedDropActions=self.supportedDropActions()
                               )
        new_list._other_data = list(self._other_data)
        new_list.extend(iterable)
        return new_list

    def __iadd__(self, iterable):
        self.extend(iterable)

    def __delitem__(self, s):
        if isinstance(s, slice):
            start, stop, step = s.indices(len(self))
            start, stop, step = _as_contiguous_range(start, stop, step)
            self.beginRemoveRows(QModelIndex(), start, stop - 1)
        else:
            s = len(self) + s if s < 0 else s
            self.beginRemoveRows(QModelIndex(), s, s)
        del self._list[s]
        del self._other_data[s]
        self.endRemoveRows()

    def __setitem__(self, s, value):
        if isinstance(s, slice):
            start, stop, step = s.indices(len(self))
            start, stop, step = _as_contiguous_range(start, stop, step)
            self.__delitem__(slice(start, stop, step))

            if not isinstance(value, list):
                value = list(value)
            self.beginInsertRows(QModelIndex(), start, start + len(value) - 1)
            self._list[s] = value
            self._other_data[s] = (_store() for _ in value)
            self.endInsertRows()
        else:
            s = len(self) + s if s < 0 else s
            self._list[s] = value
            self._other_data[s] = _store()
            self.dataChanged.emit(self.index(s), self.index(s))

    def reverse(self):
        self._list.reverse()
        self._other_data.reverse()
        self.dataChanged.emit(self.index(0), self.index(len(self) - 1))

    def sort(self, *args, **kwargs):
        indices = _argsort(self._list, *args, **kwargs)
        lst = [self._list[i] for i in indices]
        other = [self._other_data[i] for i in indices]
        for i, new_l, new_o in enumerate(zip(lst, other)):
            self._list[i] = new_l
            self._other_data[i] = new_o
        self.dataChanged.emit(self.index(0), self.index(len(self) - 1))

    def __repr__(self):
        return "PyListModel(%s)" % repr(self._list)

    def __bool__(self):
        return len(self) != 0

    def emitDataChanged(self, indexList):
        if isinstance(indexList, int):
            indexList = [indexList]

        #TODO: group indexes into ranges
        for ind in indexList:
            self.dataChanged.emit(self.index(ind), self.index(ind))

    ###########
    # Drag/drop
    ###########

    def supportedDropActions(self):
        return self._supportedDropActions

    def mimeTypes(self):
        return self.MIME_TYPES + list(QAbstractListModel.mimeTypes(self))

    def mimeData(self, indexlist):
        if len(indexlist) <= 0:
            return None

        items = [self[i.row()] for i in indexlist]
        mime = QAbstractListModel.mimeData(self, indexlist)
        data = pickle.dumps(vars)
        mime.set_data(self.MIME_TYPE, QByteArray(data))
        mime._items = items
        return mime

    def dropMimeData(self, mime, action, row, column, parent):
        if action == Qt.IgnoreAction:
            return True

        if not mime.hasFormat(self.MIME_TYPE):
            return False

        if hasattr(mime, "_vars"):
            vars_ = mime._vars
        else:
            desc = str(mime.data(self.MIME_TYPE))
            vars_ = pickle.loads(desc)

        return QAbstractListModel.dropMimeData(
            self, mime, action, row, column, parent)


class VariableListModel(PyListModel):

    MIME_TYPE = "application/x-Orange-VariableList"

    def data(self, index, role=Qt.DisplayRole):
        if self._is_index_valid_for(index, self):
            var = self[index.row()]
            if not isinstance(var, Variable):
                return super().data(index, role)
            elif role == Qt.DisplayRole:
                return var.name
            elif role == Qt.DecorationRole:
                return gui.attributeIconDict[var]
            elif role == Qt.ToolTipRole:
                return self.variable_tooltip(var)
            else:
                return PyListModel.data(self, index, role)

    def variable_tooltip(self, var):
        if isinstance(var, DiscreteVariable):
            return self.discrete_variable_tooltip(var)
        elif isinstance(var, ContinuousVariable):
            return self.continuous_variable_toltip(var)
        elif isinstance(var, StringVariable):
            return self.string_variable_tooltip(var)

    def variable_labels_tooltip(self, var):
        text = ""
        if var.attributes:
            items = [(safe_text(key), safe_text(value))
                     for key, value in var.attributes.items()]
            labels = list(map("%s = %s".__mod__, items))
            text += "<br/>Variable Labels:<br/>"
            text += "<br/>".join(labels)
        return text

    def discrete_variable_tooltip(self, var):
        text = "<b>%s</b><br/>Discrete with %i values: " %\
               (safe_text(var.name), len(var.values))
        text += ", ".join("%r" % safe_text(v) for v in var.values)
        text += self.variable_labels_tooltip(var)
        return text

    def continuous_variable_toltip(self, var):
        text = "<b>%s</b><br/>Continuous" % safe_text(var.name)
        text += self.variable_labels_tooltip(var)
        return text

    def string_variable_tooltip(self, var):
        text = "<b>%s</b><br/>String" % safe_text(var.name)
        text += self.variable_labels_tooltip(var)
        return text

    def python_variable_tooltip(self, var):
        text = "<b>%s</b><br/>Python" % safe_text(var.name)
        text += self.variable_labels_tooltip(var)
        return text

_html_replace = [("<", "&lt;"), (">", "&gt;")]


def safe_text(text):
    for old, new in _html_replace:
        text = text.replace(old, new)
    return text


class ListSingleSelectionModel(QItemSelectionModel):
    """ Item selection model for list item models with single selection.

    Defines signal:
        - selectedIndexChanged(QModelIndex)

    """
    selectedIndexChanged = Signal(QModelIndex)

    def __init__(self, model, parent=None):
        QItemSelectionModel.__init__(self, model, parent)
        self.selectionChanged.connect(self.onSelectionChanged)

    def onSelectionChanged(self, new, _):
        index = list(new.indexes())
        if index:
            index = index.pop()
        else:
            index = QModelIndex()

        self.selectedIndexChanged.emit(index)

    def selectedRow(self):
        """ Return QModelIndex of the selected row or invalid if no selection.
        """
        rows = self.selectedRows()
        if rows:
            return rows[0]
        else:
            return QModelIndex()

    def select(self, index, flags=QItemSelectionModel.ClearAndSelect):
        if isinstance(index, int):
            index = self.model().index(index)
        return QItemSelectionModel.select(self, index, flags)


def select_row(view, row):
    """
    Select a `row` in an item view.
    """
    selmodel = view.selectionModel()
    selmodel.select(view.model().index(row, 0),
                    QItemSelectionModel.ClearAndSelect |
                    QItemSelectionModel.Rows)


class ModelActionsWidget(QWidget):
    def __init__(self, actions=None, parent=None,
                 direction=QBoxLayout.LeftToRight):
        QWidget.__init__(self, parent)
        self.actions = []
        self.buttons = []
        layout = QBoxLayout(direction)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        if actions is not None:
            for action in actions:
                self.addAction(action)
        self.setLayout(layout)

    def actionButton(self, action):
        if isinstance(action, QAction):
            button = QToolButton(self)
            button.setDefaultAction(action)
            return button
        elif isinstance(action, QAbstractButton):
            return action

    def insertAction(self, ind, action, *args):
        button = self.actionButton(action)
        self.layout().insertWidget(ind, button, *args)
        self.buttons.insert(ind, button)
        self.actions.insert(ind, action)
        return button

    def addAction(self, action, *args):
        return self.insertAction(-1, action, *args)



import os, numpy

try:
    from PyQt5 import QtCore, QtWidgets, QtGui
except:
    pass

class TriggerOut:
    def __init__(self, new_object=False, additional_parameters={}):
        super().__init__()

        self.new_object = new_object

        self.__additional_parameters=additional_parameters

    def has_additional_parameter(self, name):
        return name in self.__additional_parameters.keys()

    def get_additional_parameter(self, name):
        return self.__additional_parameters[name]

class TriggerIn:
    def __init__(self, new_object=False, interrupt=False, additional_parameters={}):
        super().__init__()

        self.new_object = new_object
        self.interrupt = interrupt

        self.__additional_parameters=additional_parameters

    def has_additional_parameter(self, name):
        return name in self.__additional_parameters.keys()

    def get_additional_parameter(self, name):
        return self.__additional_parameters[name]


class TTYGrabber:
    def __init__(self,  tmpFileName = 'out.tmp.dat'):
        self.tmpFileName = tmpFileName
        self.ttyData = []
        self.outfile = False
        self.save = False

    def start(self):
        self.outfile = os.open(self.tmpFileName, os.O_RDWR|os.O_CREAT)
        self.save = os.dup(1)
        os.dup2(self.outfile, 1)
        return

    def stop(self):
        if not self.save:
            return
        os.dup2(self.save, 1)
        tmpFile = open(self.tmpFileName, "r")
        self.ttyData = tmpFile.readlines()
        tmpFile.close()
        os.close(self.outfile)
        os.remove(self.tmpFileName)

try:
    class EmittingStream(QtCore.QObject):
        textWritten = QtCore.pyqtSignal(str)

        def write(self, text):
            self.textWritten.emit(str(text))

        def flush(self):
            pass
except:
    pass

import h5py, time

subgroup_name = "surface_file"

def read_surface_file(file_name):
    if not os.path.isfile(file_name): raise ValueError("File " + file_name + " not existing")

    file = h5py.File(file_name, 'r')
    xx = file[subgroup_name + "/X"][()]
    yy = file[subgroup_name + "/Y"][()]
    zz = file[subgroup_name + "/Z"][()]

    return xx, yy, zz

def write_surface_file(zz, xx, yy, file_name, overwrite=True):

    if (os.path.isfile(file_name)) and (overwrite==True): os.remove(file_name)

    if not os.path.isfile(file_name):  # if file doesn't exist, create it.
        file = h5py.File(file_name, 'w')
        # points to the default data to be plotted
        file.attrs['default']          = subgroup_name
        # give the HDF5 root some more attributes
        file.attrs['file_name']        = file_name
        file.attrs['file_time']        = time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime())
        file.attrs['creator']          = 'write_surface_file'
        file.attrs['code']             = 'Oasys'
        file.attrs['HDF5_Version']     = h5py.version.hdf5_version
        file.attrs['h5py_version']     = h5py.version.version
        file.close()

    file = h5py.File(file_name, 'a')

    try:
        f1 = file.create_group(subgroup_name)
    except:
        f1 = file[subgroup_name]

    f1z = f1.create_dataset("Z", data=zz)
    f1x = f1.create_dataset("X", data=xx)
    f1y = f1.create_dataset("Y", data=yy)


    # NEXUS attributes for automatic plot
    f1.attrs['NX_class'] = 'NXdata'
    f1.attrs['signal'] = "Z"
    f1.attrs['axes'] = [b"Y", b"X"]

    f1z.attrs['interpretation'] = 'image'
    f1x.attrs['long_name'] = "X [m]"
    f1y.attrs['long_name'] = "Y [m]"


    file.close()

try:
    class ShowTextDialog(QtWidgets.QDialog):

        def __init__(self, title, text, width=650, height=400, parent=None, label=False, button=True):
            QtWidgets.QDialog.__init__(self, parent)
            self.setModal(True)
            self.setWindowTitle(title)
            layout = QtWidgets.QVBoxLayout(self)

            if label:
                text_area = QtWidgets.QLabel(text)
            else:
                text_edit = QtWidgets.QTextEdit("", self)
                text_edit.append(text)
                text_edit.setReadOnly(True)

                text_area = QtWidgets.QScrollArea(self)
                text_area.setWidget(text_edit)
                text_area.setWidgetResizable(False)
                text_area.setFixedHeight(height)
                text_area.setFixedWidth(width)

            layout.addWidget(text_area)

            if button:
                bbox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok)
                bbox.accepted.connect(self.accept)
                layout.addWidget(bbox)

        @classmethod
        def show_text(cls, title, text, width=650, height=400, parent=None, label=False, button=True):
            dialog = ShowTextDialog(title, text, width, height, parent, label, button)
            dialog.show()

    import sys, threading

    from PyQt5.QtWidgets import QWidget, QDialog, QVBoxLayout, QLabel
    from PyQt5.QtGui import QPainter, QPalette, QBrush, QPen, QColor
    from PyQt5.QtCore import Qt

    class ShowWaitDialog(QDialog):
        def __init__(self, title, text, width=500, height=80, parent=None):
            QDialog.__init__(self, parent)
            self.setModal(True)
            self.setWindowTitle(title)
            layout = QVBoxLayout(self)
            self.setFixedWidth(width)
            self.setFixedHeight(height)
            label = QLabel()
            label.setFixedWidth(width*0.95)
            label.setText(text)
            label.setAlignment(Qt.AlignCenter)
            label.setStyleSheet("font: 14px")
            layout.addWidget(label)
            label = QLabel()
            label.setFixedWidth(width*0.95)
            label.setText("Please wait....")
            label.setAlignment(Qt.AlignCenter)
            label.setStyleSheet("font: bold italic 16px; color: rgb(232, 120, 32);")
            layout.addWidget(label)

    class Overlay(QWidget):

        def __init__(self, container_widget=None, target_method=None, wait=0.001):

            QWidget.__init__(self, container_widget)
            self.container_widget = container_widget
            self.target_method = target_method
            palette = QPalette(self.palette())
            palette.setColor(palette.Background, Qt.transparent)
            self.setPalette(palette)
            self.__wait = wait

        def paintEvent(self, event):
            painter = QPainter()
            painter.begin(self)
            painter.setRenderHint(QPainter.Antialiasing)
            painter.fillRect(event.rect(), QBrush(QColor(255, 255, 255, 127)))
            painter.setPen(QPen(Qt.NoPen))

            for i in range(1, 7):
                if self.position_index == i:
                    painter.setBrush(QBrush(QColor(255, 165, 0)))
                else:
                    painter.setBrush(QBrush(QColor(127, 127, 127)))
                painter.drawEllipse(
                    self.width()/2 + 30 * numpy.cos(2 * numpy.pi * i / 6.0) - 10,
                    self.height()/2 + 30 * numpy.sin(2 * numpy.pi * i / 6.0) - 10,
                    20, 20)

                time.sleep(self.__wait)

            painter.end()

        def showEvent(self, event):
            self.timer = self.startTimer(0)
            self.counter = 0
            self.position_index = 0
            if not self.target_method is None:
                t = threading.Thread(target=self.target_method)
                t.start()

        def hideEvent(self, QHideEvent):
            self.killTimer(self.timer)

        def timerEvent(self, event):
            self.counter += 1
            self.position_index += 1
            if self.position_index == 7: self.position_index = 1
            self.update()
except:
    pass

class ElementInFormula(object):
    def __init__(self, element, atomic_number, n_atoms, molecular_weight):
        self._element = element
        self._atomic_number = atomic_number
        self._n_atoms = n_atoms
        self._molecular_weight = molecular_weight

class ChemicalFormulaParser(object):

    @classmethod
    def parse_formula(cls, formula):
        return parse(formula).getsyms()

def get_fwhm(histogram, bins, ret0=None):
    fwhm = ret0
    quote = ret0
    coordinates = None

    if histogram.size > 1:
        quote = numpy.max(histogram)*0.5
        cursor = numpy.where(histogram >= quote)

        if histogram[cursor].size > 1:
            bin_size    = bins[1]-bins[0]
            fwhm        = bin_size*(cursor[0][-1]-cursor[0][0])
            coordinates = (bins[cursor[0][0]], bins[cursor[0][-1]])

    return fwhm, quote, coordinates

def get_sigma(histogram, bins, ret0=None):
    if histogram.size > 1:
        frequency = histogram/numpy.sum(histogram)
        average   = numpy.sum(frequency*bins)
        return numpy.sqrt(numpy.sum(frequency*((bins-average)**2)))
    else:
        return ret0

def get_rms(histogram, bins, ret0=None):
    if histogram.size > 1:
        frequency = histogram/numpy.sum(histogram)
        return numpy.sqrt(numpy.sum(frequency*(bins**2)))
    else:
        return ret0

def get_average(histogram, bins, ret0=None):
    if histogram.size > 1:
        frequency = histogram/numpy.sum(histogram)
        return numpy.sum(frequency*bins)
    else:
        return ret0

###################################


# symbol, name, atomic number, molecular weight
_data = r"""'Ac', 'Actinium', 89, 227
'Ag', 'Silver', 47, 107.868
'Al', 'Aluminum', 13, 26.98154
'Am', 'Americium', 95, 243
'Ar', 'Argon', 18, 39.948
'As', 'Arsenic', 33, 74.9216
'At', 'Astatine', 85, 210
'Au', 'Gold', 79, 196.9665
'B', 'Boron', 5, 10.81
'Ba', 'Barium', 56, 137.33
'Be', 'Beryllium', 4, 9.01218
'Bi', 'Bismuth', 83, 208.9804
'Bk', 'Berkelium', 97, 247
'Br', 'Bromine', 35, 79.904
'C', 'Carbon', 6, 12.011
'Ca', 'Calcium', 20, 40.08
'Cd', 'Cadmium', 48, 112.41
'Ce', 'Cerium', 58, 140.12
'Cf', 'Californium', 98, 251
'Cl', 'Chlorine', 17, 35.453
'Cm', 'Curium', 96, 247
'Co', 'Cobalt', 27, 58.9332
'Cr', 'Chromium', 24, 51.996
'Cs', 'Cesium', 55, 132.9054
'Cu', 'Copper', 29, 63.546
'Dy', 'Dysprosium', 66, 162.50
'Er', 'Erbium', 68, 167.26
'Es', 'Einsteinium', 99, 254
'Eu', 'Europium', 63, 151.96
'F', 'Fluorine', 9, 18.998403
'Fe', 'Iron', 26, 55.847
'Fm', 'Fermium', 100, 257
'Fr', 'Francium', 87, 223
'Ga', 'Gallium', 31, 69.735
'Gd', 'Gadolinium', 64, 157.25
'Ge', 'Germanium', 32, 72.59
'H', 'Hydrogen', 1, 1.0079
'He', 'Helium', 2, 4.0026
'Hf', 'Hafnium', 72, 178.49
'Hg', 'Mercury', 80, 200.59
'Ho', 'Holmium', 67, 164.9304
'I', 'Iodine', 53, 126.9045
'In', 'Indium', 49, 114.82
'Ir', 'Iridium', 77, 192.22
'K', 'Potassium', 19, 39.0983
'Kr', 'Krypton', 36, 83.80
'La', 'Lanthanum', 57, 138.9055
'Li', 'Lithium', 3, 6.94
'Lr', 'Lawrencium', 103, 260
'Lu', 'Lutetium', 71, 174.96
'Md', 'Mendelevium', 101, 258
'Mg', 'Magnesium', 12, 24.305
'Mn', 'Manganese', 25, 54.9380
'Mo', 'Molybdenum', 42, 95.94
'N', 'Nitrogen', 7, 14.0067
'Na', 'Sodium', 11, 22.98977
'Nb', 'Niobium', 41, 92.9064
'Nd', 'Neodymium', 60, 144.24
'Ne', 'Neon', 10, 20.17
'Ni', 'Nickel', 28, 58.71
'No', 'Nobelium', 102, 259
'Np', 'Neptunium', 93, 237.0482
'O', 'Oxygen', 8, 15.9994
'Os', 'Osmium', 76, 190.2
'P', 'Phosphorous', 15, 30.97376
'Pa', 'Proactinium', 91, 231.0359
'Pb', 'Lead', 82, 207.2
'Pd', 'Palladium', 46, 106.4
'Pm', 'Promethium', 61, 145
'Po', 'Polonium', 84, 209
'Pr', 'Praseodymium', 59, 140.9077
'Pt', 'Platinum', 78, 195.09
'Pu', 'Plutonium', 94, 244
'Ra', 'Radium', 88, 226.0254
'Rb', 'Rubidium', 37, 85.467
'Re', 'Rhenium', 75, 186.207
'Rh', 'Rhodium', 45, 102.9055
'Rn', 'Radon', 86, 222
'Ru', 'Ruthenium', 44, 101.07
'S', 'Sulfur', 16, 32.06
'Sb', 'Antimony', 51, 121.75
'Sc', 'Scandium', 21, 44.9559
'Se', 'Selenium', 34, 78.96
'Si', 'Silicon', 14, 28.0855
'Sm', 'Samarium', 62, 150.4
'Sn', 'Tin', 50, 118.69
'Sr', 'Strontium', 38, 87.62
'Ta', 'Tantalum', 73, 180.947
'Tb', 'Terbium', 65, 158.9254
'Tc', 'Technetium', 43, 98.9062
'Te', 'Tellurium', 52, 127.60
'Th', 'Thorium', 90, 232.0381
'Ti', 'Titanium', 22, 47.90
'Tl', 'Thallium', 81, 204.37
'Tm', 'Thulium', 69, 168.9342
'U', 'Uranium', 92, 238.029
'Unh', 'Unnilhexium', 106, 263
'Unp', 'Unnilpentium', 105, 260
'Unq', 'Unnilquadium', 104, 260
'Uns', 'Unnilseptium', 107, 262
'V', 'Vanadium', 23, 50.9415
'W', 'Tungsten', 74, 183.85
'Xe', 'Xenon', 54, 131.30
'Y', 'Yttrium', 39, 88.9059
'Yb', 'Ytterbium', 70, 173.04
'Zn', 'Zinc', 30, 65.38
'Zr', 'Zirconium', 40, 91.22"""

class Element:
    def __init__(self, symbol, name, atomicnumber, molweight):
        self.sym = symbol
        self.name = name
        self.ano = atomicnumber
        self.mw = molweight

    def getweight(self):
        return self.mw

    def addsyms(self, weight, result):
        result[self.sym] = result.get(self.sym, 0) + weight

def build_dict(s):
    answer = {}
    for line in s.split("\n"):
        symbol, name, num, weight = eval(line)
        answer[symbol] = Element(symbol, name, num, weight)
    return answer

sym2elt = build_dict(_data)
del _data

class ElementSequence:
    def __init__(self, *seq):
        self.seq = list(seq)
        self.count = 1

    def append(self, thing):
        self.seq.append(thing)

    def getweight(self):
        sum = 0.0
        for thing in self.seq:
            sum = sum + thing.getweight()
        return sum * self.count

    def set_count(self, n):
        self.count = n

    def __len__(self):
        return len(self.seq)

    def addsyms(self, weight, result):
        totalweight = weight * self.count
        for thing in self.seq:
            thing.addsyms(totalweight, result)

    def displaysyms(self):
        result = {}
        self.addsyms(1, result)
        #items = sorted(result.items())

        for sym, count in result.items():
            print (sym, "(" + sym2elt[sym].name + "):", count)

    def getsyms(self):
        result = {}
        self.addsyms(1, result)
        items = sorted(result.items())

        result = []
        for sym, count in items:
            result.append (ElementInFormula(sym, sym2elt[sym].ano, count, sym2elt[sym].mw))

        return result

NAME, NUM, LPAREN, RPAREN, EOS = range(5)
import re
_lexer = re.compile(r"[A-Z][a-z]*|\d+|[()]|<EOS>").match
del re

class Tokenizer:
    def __init__(self, input):
        self.input = input + "<EOS>"
        self.i = 0

    def gettoken(self):
        global ttype, tvalue
        self.lasti = self.i
        m = _lexer(self.input, self.i)
        if m is None:
            self.error("unexpected character")
        self.i = m.end()
        tvalue = m.group()
        if tvalue == "(":
            ttype = LPAREN
        elif tvalue == ")":
            ttype = RPAREN
        elif tvalue == "<EOS>":
            ttype = EOS
        elif "0" <= tvalue[0] <= "9":
            ttype = NUM
            tvalue = int(tvalue)
        else:
            ttype = NAME

    def error(self, msg):
        emsg = msg + ":\n"
        emsg = emsg + self.input[:-5] + "\n"  # strip <EOS>
        emsg = emsg + " " * self.lasti + "^\n"

        raise ValueError(emsg)

def parse(s):
    global t, ttype, tvalue
    t = Tokenizer(s)
    t.gettoken()
    seq = parse_sequence()
    if ttype != EOS:
        t.error("expected end of input")
    return seq

def parse_sequence():
    global t, ttype, tvalue
    seq = ElementSequence()
    while ttype in (LPAREN, NAME):
        # parenthesized expression or straight name
        if ttype == LPAREN:
            t.gettoken()
            thisguy = parse_sequence()
            if ttype != RPAREN:
                t.error("expected right paren")
            t.gettoken()
        else:
            assert ttype == NAME
            if tvalue in sym2elt:
                thisguy = ElementSequence(sym2elt[tvalue])
            else:
                t.error("'" + tvalue + "' is not an element symbol")
            t.gettoken()
        # followed by optional count
        if ttype == NUM:
            thisguy.set_count(tvalue)
            t.gettoken()
        seq.append(thisguy)
    if len(seq) == 0:
        t.error("empty sequence")
    return seq






#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ----------------------------------------------------------------------- #
# Copyright (c) 2021, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2021. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# ----------------------------------------------------------------------- #
import numpy
from srxraylib.metrology import profiles_simulation, dabam

class ErrorProfileInputParameters():
    def __init__(self, widget=None):
        if not widget is None:
            self.si_to_user_units              = widget.si_to_user_units
            self.kind_of_profile_x             = widget.kind_of_profile_x
            self.kind_of_profile_y             = widget.kind_of_profile_y
            self.step_x                        = widget.step_x
            self.step_y                        = widget.step_y
            self.dimension_x                   = widget.dimension_x
            self.dimension_y                   = widget.dimension_y
            self.power_law_exponent_beta_x     = widget.power_law_exponent_beta_x
            self.power_law_exponent_beta_y     = widget.power_law_exponent_beta_y
            self.correlation_length_x          = widget.correlation_length_x
            self.correlation_length_y          = widget.correlation_length_y
            self.error_type_x                  = widget.error_type_x
            self.error_type_y                  = widget.error_type_y
            self.rms_x                         = widget.rms_x
            self.montecarlo_seed_x             = widget.montecarlo_seed_x
            self.rms_y                         = widget.rms_y
            self.montecarlo_seed_y             = widget.montecarlo_seed_y
            self.heigth_profile_1D_file_name_x = widget.heigth_profile_1D_file_name_x
            self.delimiter_x                   = widget.delimiter_x
            self.conversion_factor_x_x         = widget.conversion_factor_x_x
            self.conversion_factor_x_y         = widget.conversion_factor_x_y
            self.center_x                      = widget.center_x
            self.modify_x                      = widget.modify_x
            self.new_length_x                  = widget.new_length_x
            self.filler_value_x                = widget.filler_value_x
            self.renormalize_x                 = widget.renormalize_x
            self.heigth_profile_1D_file_name_y = widget.heigth_profile_1D_file_name_y
            self.delimiter_y                   = widget.delimiter_y
            self.conversion_factor_y_x         = widget.conversion_factor_y_x
            self.conversion_factor_y_y         = widget.conversion_factor_y_y
            self.center_y                      = widget.center_y
            self.modify_y                      = widget.modify_y
            self.new_length_y                  = widget.new_length_y
            self.filler_value_y                = widget.filler_value_y
            self.renormalize_y                 = widget.renormalize_y
        else:
            self.si_to_user_units = 1e-3
            self.kind_of_profile_x = 0
            self.kind_of_profile_y = 0
            self.step_x = 0.0
            self.step_y = 0.0
            self.dimension_x = 0.0
            self.dimension_y = 0.0
            self.power_law_exponent_beta_x = 3.0
            self.power_law_exponent_beta_y = 3.0
            self.correlation_length_x = 0.3
            self.correlation_length_y = 0.3
            self.error_type_x = profiles_simulation.FIGURE_ERROR
            self.error_type_y = profiles_simulation.FIGURE_ERROR
            self.rms_x = 0.1
            self.montecarlo_seed_x = 8787
            self.rms_y = 1
            self.montecarlo_seed_y = 8788
            self.heigth_profile_1D_file_name_x = "mirror_1D_x.dat"
            self.delimiter_x = 0
            self.conversion_factor_x_x = 0.001
            self.conversion_factor_x_y = 1e-6
            self.center_x = 1
            self.modify_x = 0
            self.new_length_x = 0.0
            self.filler_value_x = 0.0
            self.renormalize_x = 0
            self.heigth_profile_1D_file_name_y = "mirror_1D_y.dat"
            self.delimiter_y = 0
            self.conversion_factor_y_x = 0.001
            self.conversion_factor_y_y = 1e-6
            self.center_y = 1
            self.modify_y = 0
            self.new_length_y = 0.0
            self.filler_value_y = 0.0
            self.renormalize_y = 0
            
class DabamInputParameters():


    def __init__(self, dabam_server, widget=None):
        if not widget is None:
            self.server                        = dabam_server
            self.si_to_user_units              = widget.si_to_user_units
            self.use_undetrended               = widget.use_undetrended
            self.step_x                        = widget.step_x
            self.dimension_x                   = widget.dimension_x
            self.kind_of_profile_x             = widget.kind_of_profile_x
            self.power_law_exponent_beta_x     = widget.power_law_exponent_beta_x
            self.correlation_length_x          = widget.correlation_length_x
            self.error_type_x                  = widget.error_type_x
            self.rms_x                         = widget.rms_x
            self.montecarlo_seed_x             = widget.montecarlo_seed_x
            self.heigth_profile_1D_file_name_x = widget.heigth_profile_1D_file_name_x
            self.delimiter_x                   = widget.delimiter_x
            self.conversion_factor_x_x         = widget.conversion_factor_x_x
            self.conversion_factor_x_y         = widget.conversion_factor_x_y
            self.center_x                      = widget.center_x
            self.modify_x                      = widget.modify_x
            self.new_length_x                  = widget.new_length_x
            self.filler_value_x                = widget.filler_value_x
            self.renormalize_x                 = widget.renormalize_x
            self.center_y                      = widget.center_y
            self.modify_y                      = widget.modify_y
            self.new_length_y                  = widget.new_length_y
            self.filler_value_y                = widget.filler_value_y
            self.renormalize_y                 = widget.renormalize_y
            self.error_type_y                  = widget.error_type_y
            self.rms_y                         = widget.rms_y
        else:
            self.server = dabam_server
            self.si_to_user_units = 1e-3
            self.use_undetrended = 0
            self.step_x = 0.0
            self.dimension_x = 0.0
            self.kind_of_profile_x = 3
            self.power_law_exponent_beta_x = 3.0
            self.correlation_length_x = 0.3
            self.error_type_x = profiles_simulation.FIGURE_ERROR
            self.rms_x = 0.1
            self.montecarlo_seed_x = 8787
            self.heigth_profile_1D_file_name_x = "mirror_1D_x.dat"
            self.delimiter_x = 0
            self.conversion_factor_x_x = 0.001
            self.conversion_factor_x_y = 1e-6
            self.center_x = 1
            self.modify_x = 0
            self.new_length_x = 0.0
            self.filler_value_x = 0.0
            self.renormalize_x = 0
            self.center_y = 1
            self.modify_y = 0
            self.new_length_y = 0.0
            self.filler_value_y = 0.0
            self.renormalize_y = 1
            self.error_type_y = 0
            self.rms_y = 0.9

def calculate_heigth_profile(input_parameters):
    #### LENGTH
    if input_parameters.kind_of_profile_y == 2:
        combination = "E"

        if input_parameters.delimiter_y == 1:
            profile_1D_y_x, profile_1D_y_y = numpy.loadtxt(input_parameters.heigth_profile_1D_file_name_y, delimiter='\t', unpack=True)
        else:
            profile_1D_y_x, profile_1D_y_y = numpy.loadtxt(input_parameters.heigth_profile_1D_file_name_y, unpack=True)

        profile_1D_y_x *= input_parameters.conversion_factor_y_x
        profile_1D_y_y *= input_parameters.conversion_factor_y_y

        first_coord = profile_1D_y_x[0]
        second_coord = profile_1D_y_x[1]
        last_coord = profile_1D_y_x[-1]
        step = numpy.abs(second_coord - first_coord)
        length = numpy.abs(last_coord - first_coord)
        n_points_old = len(profile_1D_y_x)

        if input_parameters.modify_y == 2:
            profile_1D_y_x_temp = profile_1D_y_x
            profile_1D_y_y_temp = profile_1D_y_y

            if input_parameters.new_length_y > length:
                difference = input_parameters.new_length_y - length

                n_added_points = int(difference / step)
                if difference % step == 0:
                    n_added_points += 1
                if n_added_points % 2 != 0:
                    n_added_points += 1

                profile_1D_y_x = numpy.arange(n_added_points + n_points_old) * step
                profile_1D_y_y = numpy.ones(n_added_points + n_points_old) * input_parameters.filler_value_y * 1e-9 * input_parameters.si_to_user_units
                profile_1D_y_y[int(n_added_points / 2): n_points_old + int(n_added_points / 2)] = profile_1D_y_y_temp
            elif input_parameters.new_length_y < length:
                difference = length - input_parameters.new_length_y

                n_removed_points = int(difference / step)
                if difference % step == 0:
                    n_removed_points -= 1
                if n_removed_points % 2 != 0:
                    n_removed_points -= 1

                if n_removed_points >= 2:
                    profile_1D_y_x = profile_1D_y_x_temp[0: (n_points_old - n_removed_points)]
                    profile_1D_y_y = profile_1D_y_y_temp[(int(n_removed_points / 2) - 1): (n_points_old - int(n_removed_points / 2) - 1)]

                else:
                    profile_1D_y_x = profile_1D_y_x_temp
                    profile_1D_y_y = profile_1D_y_y_temp
            else:
                profile_1D_y_x = profile_1D_y_x_temp
                profile_1D_y_y = profile_1D_y_y_temp

        elif input_parameters.modify_y == 1:
            scale_factor_y = input_parameters.new_length_y / length
            profile_1D_y_x *= scale_factor_y

        if input_parameters.center_y:
            first_coord = profile_1D_y_x[0]
            last_coord = profile_1D_y_x[-1]
            length = numpy.abs(last_coord - first_coord)

            profile_1D_y_x_temp = numpy.linspace(-length / 2, length / 2, len(profile_1D_y_x))
            profile_1D_y_x = profile_1D_y_x_temp

        if input_parameters.renormalize_y == 0:
            rms_y = None
        else:
            if input_parameters.error_type_y == profiles_simulation.FIGURE_ERROR:
                rms_y = input_parameters.rms_y * 1e-9 * input_parameters.si_to_user_units  # from nm to m
            else:
                rms_y = input_parameters.rms_y * 1e-6  # from urad to rad
    else:
        if input_parameters.kind_of_profile_y == 0:
            combination = "F"
        else:
            combination = "G"

        profile_1D_y_x = None
        profile_1D_y_y = None

        if input_parameters.error_type_y == profiles_simulation.FIGURE_ERROR:
            rms_y = input_parameters.rms_y * 1e-9 * input_parameters.si_to_user_units  # from nm to m
        else:
            rms_y = input_parameters.rms_y * 1e-6  # from urad to rad

    #### WIDTH
    if input_parameters.kind_of_profile_x == 2:
        combination += "E"

        if input_parameters.delimiter_x == 1:
            profile_1D_x_x, profile_1D_x_y = numpy.loadtxt(input_parameters.heigth_profile_1D_file_name_x, delimiter='\t', unpack=True)
        else:
            profile_1D_x_x, profile_1D_x_y = numpy.loadtxt(input_parameters.heigth_profile_1D_file_name_x, unpack=True)

        profile_1D_x_x *= input_parameters.conversion_factor_x_x
        profile_1D_x_y *= input_parameters.conversion_factor_x_y

        first_coord = profile_1D_x_x[0]
        second_coord = profile_1D_x_x[1]
        last_coord = profile_1D_x_x[-1]
        step = numpy.abs(second_coord - first_coord)
        length = numpy.abs(last_coord - first_coord)
        n_points_old = len(profile_1D_x_x)

        if input_parameters.modify_x == 2:
            profile_1D_x_x_temp = profile_1D_x_x
            profile_1D_x_y_temp = profile_1D_x_y

            if input_parameters.new_length_x > length:
                difference = input_parameters.new_length_x - length

                n_added_points = int(difference / step)
                if difference % step == 0:
                    n_added_points += 1
                if n_added_points % 2 != 0:
                    n_added_points += 1

                profile_1D_x_x = numpy.arange(n_added_points + n_points_old) * step
                profile_1D_x_y = numpy.ones(n_added_points + n_points_old) * input_parameters.filler_value_x * 1e-9 * input_parameters.si_to_user_units
                profile_1D_x_y[int(n_added_points / 2): n_points_old + int(n_added_points / 2)] = profile_1D_x_y_temp
            elif input_parameters.new_length_x < length:
                difference = length - input_parameters.new_length_x

                n_removed_points = int(difference / step)
                if difference % step == 0:
                    n_removed_points -= 1
                if n_removed_points % 2 != 0:
                    n_removed_points -= 1

                if n_removed_points >= 2:
                    profile_1D_x_x = profile_1D_x_x_temp[0: (n_points_old - n_removed_points)]
                    profile_1D_x_y = profile_1D_x_y_temp[(int(n_removed_points / 2) - 1): (n_points_old - int(n_removed_points / 2) - 1)]

                else:
                    profile_1D_x_x = profile_1D_x_x_temp
                    profile_1D_x_y = profile_1D_x_y_temp
            else:
                profile_1D_x_x = profile_1D_x_x_temp
                profile_1D_x_y = profile_1D_x_y_temp

        elif input_parameters.modify_x == 1:
            scale_factor_x = input_parameters.new_length_x / length
            profile_1D_x_x *= scale_factor_x

        if input_parameters.center_x:
            first_coord = profile_1D_x_x[0]
            last_coord = profile_1D_x_x[-1]
            length = numpy.abs(last_coord - first_coord)

            profile_1D_x_x_temp = numpy.linspace(-length / 2, length / 2, len(profile_1D_x_x))
            profile_1D_x_x = profile_1D_x_x_temp

        if input_parameters.renormalize_x == 0:
            rms_x = None
        else:
            if input_parameters.error_type_x == profiles_simulation.FIGURE_ERROR:
                rms_x = input_parameters.rms_x * 1e-9 * input_parameters.si_to_user_units  # from nm to m
            else:
                rms_x = input_parameters.rms_x * 1e-6  # from urad to rad

    else:
        profile_1D_x_x = None
        profile_1D_x_y = None

        if input_parameters.kind_of_profile_x == 0:
            combination += "F"
        else:
            combination += "G"

        if input_parameters.error_type_x == profiles_simulation.FIGURE_ERROR:
            rms_x = input_parameters.rms_x * 1e-9 * input_parameters.si_to_user_units  # from nm to m
        else:
            rms_x = input_parameters.rms_x * 1e-6  # from urad to rad

    xx, yy, zz = profiles_simulation.simulate_profile_2D(combination=combination,
                                                         mirror_length=input_parameters.dimension_y,
                                                         step_l=input_parameters.step_y,
                                                         random_seed_l=input_parameters.montecarlo_seed_y,
                                                         error_type_l=input_parameters.error_type_y,
                                                         rms_l=rms_y,
                                                         power_law_exponent_beta_l=input_parameters.power_law_exponent_beta_y,
                                                         correlation_length_l=input_parameters.correlation_length_y,
                                                         x_l=profile_1D_y_x,
                                                         y_l=profile_1D_y_y,
                                                         mirror_width=input_parameters.dimension_x,
                                                         step_w=input_parameters.step_x,
                                                         random_seed_w=input_parameters.montecarlo_seed_x,
                                                         error_type_w=input_parameters.error_type_x,
                                                         rms_w=rms_x,
                                                         power_law_exponent_beta_w=input_parameters.power_law_exponent_beta_x,
                                                         correlation_length_w=input_parameters.correlation_length_x,
                                                         x_w=profile_1D_x_x,
                                                         y_w=profile_1D_x_y)

    return xx, yy, zz

def calculate_dabam_profile(input_parameters):
    combination = "E"

    #### LENGTH
    if input_parameters.modify_y == 2:
        profile_1D_y_x_temp = input_parameters.si_to_user_units * input_parameters.server.y
        if input_parameters.use_undetrended == 0:
            profile_1D_y_y_temp = input_parameters.si_to_user_units * input_parameters.server.zHeights
        else:
            profile_1D_y_y_temp = input_parameters.si_to_user_units * input_parameters.server.zHeightsUndetrended

        first_coord = profile_1D_y_x_temp[0]
        second_coord = profile_1D_y_x_temp[1]
        last_coord = profile_1D_y_x_temp[-1]
        step = numpy.abs(second_coord - first_coord)
        length = numpy.abs(last_coord - first_coord)
        n_points_old = len(profile_1D_y_x_temp)

        if input_parameters.new_length_y > length:
            difference = input_parameters.new_length_y - length

            n_added_points = int(difference / step)
            if difference % step == 0:
                n_added_points += 1
            if n_added_points % 2 != 0:
                n_added_points += 1

            profile_1D_y_x = numpy.arange(n_added_points + n_points_old) * step
            profile_1D_y_y = numpy.ones(n_added_points + n_points_old) * input_parameters.filler_value_y * 1e-9 * input_parameters.si_to_user_units
            profile_1D_y_y[int(n_added_points / 2): n_points_old + int(n_added_points / 2)] = profile_1D_y_y_temp
        elif input_parameters.new_length_y < length:
            difference = length - input_parameters.new_length_y

            n_removed_points = int(difference / step)
            if difference % step == 0:
                n_removed_points -= 1
            if n_removed_points % 2 != 0:
                n_removed_points -= 1

            if n_removed_points >= 2:
                profile_1D_y_x = profile_1D_y_x_temp[0: (n_points_old - n_removed_points)]
                profile_1D_y_y = profile_1D_y_y_temp[(int(n_removed_points / 2) - 1): (n_points_old - int(n_removed_points / 2) - 1)]

            else:
                profile_1D_y_x = profile_1D_y_x_temp
                profile_1D_y_y = profile_1D_y_y_temp
        else:
            profile_1D_y_x = profile_1D_y_x_temp
            profile_1D_y_y = profile_1D_y_y_temp

    else:
        if input_parameters.modify_y == 0:
            profile_1D_y_x = input_parameters.si_to_user_units * input_parameters.server.y
        elif input_parameters.modify_y == 1:
            scale_factor_y = input_parameters.new_length_y / (input_parameters.si_to_user_units * (max(input_parameters.server.y) - min(input_parameters.server.y)))
            profile_1D_y_x = input_parameters.si_to_user_units * input_parameters.server.y * scale_factor_y

        if input_parameters.use_undetrended == 0:
            profile_1D_y_y = input_parameters.si_to_user_units * input_parameters.server.zHeights
        else:
            profile_1D_y_y = input_parameters.si_to_user_units * input_parameters.server.zHeightsUndetrended

    if input_parameters.center_y:
        first_coord = profile_1D_y_x[0]
        last_coord = profile_1D_y_x[-1]
        length = numpy.abs(last_coord - first_coord)

        profile_1D_y_x_temp = numpy.linspace(-length / 2, length / 2, len(profile_1D_y_x))

        profile_1D_y_x = profile_1D_y_x_temp

    if input_parameters.renormalize_y == 0:
        rms_y = None
    else:
        if input_parameters.error_type_y == profiles_simulation.FIGURE_ERROR:
            rms_y = input_parameters.rms_y * 1e-9 * input_parameters.si_to_user_units  # from nm to m
        else:
            rms_y = input_parameters.rms_y * 1e-6  # from urad to rad

    #### WIDTH
    if input_parameters.kind_of_profile_x == 3:
        combination += "F"

        xx, yy, zz = profiles_simulation.simulate_profile_2D(combination=combination,
                                                             error_type_l=input_parameters.error_type_y,
                                                             rms_l=rms_y,
                                                             x_l=profile_1D_y_x,
                                                             y_l=profile_1D_y_y,
                                                             mirror_width=input_parameters.dimension_x,
                                                             step_w=input_parameters.step_x,
                                                             rms_w=0.0)
    else:
        if input_parameters.kind_of_profile_x == 2:
            combination += "E"

            if input_parameters.delimiter_x == 1:
                profile_1D_x_x, profile_1D_x_y = numpy.loadtxt(input_parameters.heigth_profile_1D_file_name_x, delimiter='\t', unpack=True)
            else:
                profile_1D_x_x, profile_1D_x_y = numpy.loadtxt(input_parameters.heigth_profile_1D_file_name_x, unpack=True)

            profile_1D_x_x *= input_parameters.conversion_factor_x_x
            profile_1D_x_y *= input_parameters.conversion_factor_x_y

            first_coord = profile_1D_x_x[0]
            second_coord = profile_1D_x_x[1]
            last_coord = profile_1D_x_x[-1]
            step = numpy.abs(second_coord - first_coord)
            length = numpy.abs(last_coord - first_coord)
            n_points_old = len(profile_1D_x_x)

            if input_parameters.modify_x == 2:
                profile_1D_x_x_temp = profile_1D_x_x
                profile_1D_x_y_temp = profile_1D_x_y

                if input_parameters.new_length_x > length:
                    difference = input_parameters.new_length_x - length

                    n_added_points = int(difference / step)
                    if difference % step == 0:
                        n_added_points += 1
                    if n_added_points % 2 != 0:
                        n_added_points += 1

                    profile_1D_x_x = numpy.arange(n_added_points + n_points_old) * step
                    profile_1D_x_y = numpy.ones(n_added_points + n_points_old) * input_parameters.filler_value_x * 1e-9 * input_parameters.si_to_user_units
                    profile_1D_x_y[int(n_added_points / 2): n_points_old + int(n_added_points / 2)] = profile_1D_x_y_temp
                elif input_parameters.new_length_x < length:
                    difference = length - input_parameters.new_length_x

                    n_removed_points = int(difference / step)
                    if difference % step == 0:
                        n_removed_points -= 1
                    if n_removed_points % 2 != 0:
                        n_removed_points -= 1

                    if n_removed_points >= 2:
                        profile_1D_x_x = profile_1D_x_x_temp[0: (n_points_old - n_removed_points)]
                        profile_1D_x_y = profile_1D_x_y_temp[(int(n_removed_points / 2) - 1): (n_points_old - int(n_removed_points / 2) - 1)]

                    else:
                        profile_1D_x_x = profile_1D_x_x_temp
                        profile_1D_x_y = profile_1D_x_y_temp
                else:
                    profile_1D_x_x = profile_1D_x_x_temp
                    profile_1D_x_y = profile_1D_x_y_temp

            elif input_parameters.modify_x == 1:
                scale_factor_x = input_parameters.new_length_x / length
                profile_1D_x_x *= scale_factor_x

            if input_parameters.center_x:
                first_coord = profile_1D_x_x[0]
                last_coord = profile_1D_x_x[-1]
                length = numpy.abs(last_coord - first_coord)

                profile_1D_x_x_temp = numpy.linspace(-length / 2, length / 2, len(profile_1D_x_x))
                profile_1D_x_x = profile_1D_x_x_temp

            if input_parameters.renormalize_x == 0:
                rms_x = None
            else:
                if input_parameters.error_type_x == profiles_simulation.FIGURE_ERROR:
                    rms_x = input_parameters.rms_x * 1e-9 * input_parameters.si_to_user_units  # from nm to m
                else:
                    rms_x = input_parameters.rms_x * 1e-6  # from urad to rad

        else:
            profile_1D_x_x = None
            profile_1D_x_y = None

            if input_parameters.kind_of_profile_x == 0:
                combination += "F"
            else:
                combination += "G"

            if input_parameters.error_type_x == profiles_simulation.FIGURE_ERROR:
                rms_x = input_parameters.rms_x * 1e-9 * input_parameters.si_to_user_units  # from nm to m
            else:
                rms_x = input_parameters.rms_x * 1e-6  # from urad to rad

        xx, yy, zz = profiles_simulation.simulate_profile_2D(combination=combination,
                                                             error_type_l=input_parameters.error_type_y,
                                                             rms_l=rms_y,
                                                             x_l=profile_1D_y_x,
                                                             y_l=profile_1D_y_y,
                                                             mirror_width=input_parameters.dimension_x,
                                                             step_w=input_parameters.step_x,
                                                             random_seed_w=input_parameters.montecarlo_seed_x,
                                                             error_type_w=input_parameters.error_type_x,
                                                             rms_w=rms_x,
                                                             power_law_exponent_beta_w=input_parameters.power_law_exponent_beta_x,
                                                             correlation_length_w=input_parameters.correlation_length_x,
                                                             x_w=profile_1D_x_x,
                                                             y_w=profile_1D_x_y)

    return xx, yy, zz



import sys
import os
import logging
import itertools
import json
import concurrent.futures

from xml.sax.saxutils import escape
from distutils import version

import pkg_resources
import requests

try:
    import docutils.core
except ImportError:
    docutils = None

from PyQt5.QtWidgets import (
    QWidget, QDialog, QLabel, QLineEdit, QTreeView, QHeaderView,
    QTextBrowser, QDialogButtonBox, QProgressDialog,
    QVBoxLayout, QHBoxLayout
)

from PyQt5.QtGui import (
    QStandardItemModel, QStandardItem, QPalette, QTextOption
)

from PyQt5.QtCore import (
    QSortFilterProxyModel, QItemSelectionModel,
    Qt, QSize, QTimer, QThread)
from PyQt5.QtCore import pyqtSignal as Signal, pyqtSlot as Slot

from urllib.request import urlopen

from orangecanvas.gui.utils import message_warning, message_information, \
                        message_critical as message_error
from orangecanvas.help.manager import get_dist_meta, trim

from orangecanvas.resources import package_dirname

PYPI_API_JSON = "https://pypi.org/pypi/{name}/json"

#####################################################
# MAKE INTERNAL LIBRARIES DYNAMICAL

# read add-on list

INTERNAL_LIBRARIES = []
MAX_VERSION        = {}

if sys.version[:3]=="3.8": internal_libraries_file = "INTERNAL_LIBRARIES_PY38.txt"
else: internal_libraries_file = "INTERNAL_LIBRARIES_PY37.txt"

for a in open(os.path.join(package_dirname("oasys.application"), "data", internal_libraries_file), "rt"):
    a = a.strip()
    if a:
        a = a.split(sep="==")
        library_name = a[0].lower()
        INTERNAL_LIBRARIES.append(library_name)
        if len(a) == 2:
            MAX_VERSION[library_name] = a[1]
        else:
            MAX_VERSION[library_name] = None

# query PyPI

internal_libraries_list = []
is_auto_update = True

try:
    for package in INTERNAL_LIBRARIES:
        r = urlopen(PYPI_API_JSON.format(name=package)).read().decode("utf-8")
        p = json.loads(r)
        p["releases"] = p["releases"][p["info"]["version"]] # load only the last version

        internal_libraries_list.append(p)
except:
    is_auto_update = False


log = logging.getLogger(__name__)

from oasys.application.addons import Uninstall, Upgrade, Install
from oasys.application.addons import OSX_NSURL_toLocalFile, Installer
from oasys.application.addons import get_meta_from_archive, cleanup, TristateCheckItemDelegate, method_queued, unique
from oasys.application.addons import Installed, Installable, Available

def is_updatable(item):
    if isinstance(item, Available) or isinstance(item, Installable) or item.installable is None:
        return False
    inst, dist = item
    project_name = str(dist.project_name).lower()
    try:
        if version.StrictVersion(dist.version) < version.StrictVersion(inst.version):
            if MAX_VERSION[project_name] is None: return True
            else: return version.StrictVersion(MAX_VERSION[project_name]) >= version.StrictVersion(inst.version)
    except ValueError:
        if version.LooseVersion(dist.version) < version.LooseVersion(inst.version):
            if MAX_VERSION[project_name] is None: return True
            else: return version.LooseVersion(MAX_VERSION[project_name]) >= version.LooseVersion(inst.version)

def is_installable(item):
    if isinstance(item, Available) or isinstance(item, Installable) or item.installable is None: return True
    else: return False

class InternalLibrariesManagerWidget(QWidget):

    statechanged = Signal()

    def __init__(self, parent=None, **kwargs):
        super(InternalLibrariesManagerWidget, self).__init__(parent, **kwargs)
        self.__items = []
        self.setLayout(QVBoxLayout())

        self.__header = QLabel(
            wordWrap=True,
            textFormat=Qt.RichText
        )
        self.__search = QLineEdit(
            placeholderText=self.tr("Filter")
        )

        self.tophlayout = topline = QHBoxLayout()
        topline.addWidget(self.__search)
        self.layout().addLayout(topline)

        self.__view = view = QTreeView(
            rootIsDecorated=False,
            editTriggers=QTreeView.NoEditTriggers,
            selectionMode=QTreeView.SingleSelection,
            alternatingRowColors=True
        )
        self.__view.setItemDelegateForColumn(0, TristateCheckItemDelegate())
        self.layout().addWidget(view)

        self.__model = model = QStandardItemModel()
        model.setHorizontalHeaderLabels(["", "Name", "Version", "Action"])
        model.dataChanged.connect(self.__data_changed)
        self.__proxy = proxy = QSortFilterProxyModel(
            filterKeyColumn=1,
            filterCaseSensitivity=Qt.CaseInsensitive
        )
        proxy.setSourceModel(model)
        self.__search.textChanged.connect(proxy.setFilterFixedString)

        view.setModel(proxy)
        view.selectionModel().selectionChanged.connect(
            self.__update_details
        )
        header = self.__view.header()
        header.setSectionResizeMode(0, QHeaderView.Fixed)
        header.setSectionResizeMode(2, QHeaderView.ResizeToContents)

        self.__details = QTextBrowser(
            frameShape=QTextBrowser.NoFrame,
            readOnly=True,
            lineWrapMode=QTextBrowser.WidgetWidth,
            openExternalLinks=True,
        )

        self.__details.setWordWrapMode(QTextOption.WordWrap)
        palette = QPalette(self.palette())
        palette.setColor(QPalette.Base, Qt.transparent)
        self.__details.setPalette(palette)
        self.layout().addWidget(self.__details)

    def set_items(self, items):
        self.__items = items
        model = self.__model
        model.clear()
        model.setHorizontalHeaderLabels(["", "Name", "Version", "Action"])

        for item in items:
            if isinstance(item, Installed):
                installed = True
                ins, dist = item
                name = dist.project_name
                summary = get_dist_meta(dist).get("Summary", "")
                version = ins.version if ins is not None else dist.version
            else:
                installed = False
                (ins,) = item
                dist = None
                name = ins.name
                summary = ins.summary
                version = ins.version

            updatable = is_updatable(item)

            item1 = QStandardItem()
            item1.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable |
                           Qt.ItemIsUserCheckable |
                           (Qt.ItemIsTristate if updatable else 0))
            item1.setEnabled(False)

            item1.setCheckState(Qt.Checked)

            #if installed and updatable:
            #    item1.setCheckState(Qt.Checked)
            #elif installed:
            #    item1.setCheckState(Qt.Checked)
            #else:
            #    item1.setCheckState(Qt.Unchecked)

            item2 = QStandardItem(cleanup(name))

            item2.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)
            item2.setToolTip(summary)
            item2.setData(item, Qt.UserRole)

            if updatable:
                version = "{} < {}".format(dist.version, ins.version)

            item3 = QStandardItem(version)
            item3.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)

            item4 = QStandardItem()
            item4.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)

            if updatable:
                item4.setText("Update")
            elif not installed:
                item4.setText("Install")

            model.appendRow([item1, item2, item3, item4])

        self.__view.resizeColumnToContents(0)
        self.__view.setColumnWidth(
            1, max(150, self.__view.sizeHintForColumn(1)))
        self.__view.setColumnWidth(
            2, max(150, self.__view.sizeHintForColumn(2)))

        if self.__items:
            self.__view.selectionModel().select(
                self.__view.model().index(0, 0),
                QItemSelectionModel.Select | QItemSelectionModel.Rows
            )

    def item_state(self):
        steps = []
        for i, item in enumerate(self.__items):
            modelitem = self.__model.item(i, 0)
            state = modelitem.checkState()
            if modelitem.flags() & Qt.ItemIsTristate and state == Qt.Checked:
                steps.append((Upgrade, item))
            elif isinstance(item, Available) and state == Qt.Checked:
                steps.append((Install, item))
            elif isinstance(item, Installed) and state == Qt.Unchecked:
                steps.append((Uninstall, item))

        return steps

    def __selected_row(self):
        indices = self.__view.selectedIndexes()
        if indices:
            proxy = self.__view.model()
            indices = [proxy.mapToSource(index) for index in indices]
            return indices[0].row()
        else:
            return -1

    def set_install_projects(self, names):
        """Mark for installation the add-ons that match any of names"""
        model = self.__model
        for row in range(model.rowCount()):
            item = model.item(row, 1)
            if item.text() in names:
                model.item(row, 0).setCheckState(Qt.Checked)

    def __data_changed(self, topleft, bottomright):
        rows = range(topleft.row(), bottomright.row() + 1)
        for i in rows:
            modelitem = self.__model.item(i, 0)
            actionitem = self.__model.item(i, 3)
            item = self.__items[i]

            state = modelitem.checkState()
            flags = modelitem.flags()

            if flags & Qt.ItemIsTristate and state == Qt.Checked:
                actionitem.setText("Update")
            elif isinstance(item, Available) and state == Qt.Checked:
                actionitem.setText("Install")
            elif isinstance(item, Installed) and state == Qt.Unchecked:
                actionitem.setText("Uninstall")
            else:
                actionitem.setText("")
        self.statechanged.emit()

    def __update_details(self):
        index = self.__selected_row()
        if index == -1:
            self.__details.setText("")
        else:
            item = self.__model.item(index, 1)
            item = item.data(Qt.UserRole)
            assert isinstance(item, (Installed, Available))
            text = self._detailed_text(item)
            self.__details.setText(text)

    def _detailed_text(self, item):
        if isinstance(item, Installed):
            remote, dist = item
            if remote is None:
                meta = get_dist_meta(dist)
                description = meta.get("Description") or meta.get('Summary')
            else:
                description = remote.description
        else:
            description = item[0].description

        if docutils is not None:
            try:
                html = docutils.core.publish_string(
                    trim(description),
                    writer_name="html",
                    settings_overrides={
                        "output-encoding": "utf-8",
                        # "embed-stylesheet": False,
                        # "stylesheet": [],
                        # "stylesheet_path": []
                    }
                ).decode("utf-8")

            except docutils.utils.SystemMessage:
                html = "<pre>{}<pre>".format(escape(description))
            except Exception:
                html = "<pre>{}<pre>".format(escape(description))
        else:
            html = "<pre>{}<pre>".format(escape(description))
        return html

    def sizeHint(self):
        return QSize(480, 420)

class InternalLibrariesManagerDialog(QDialog):
    _packages = None

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, acceptDrops=True, **kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self.internallibrarieswidget = InternalLibrariesManagerWidget()
        self.layout().addWidget(self.internallibrarieswidget)

        info_bar = QWidget()
        info_layout = QHBoxLayout()
        info_bar.setLayout(info_layout)
        self.layout().addWidget(info_bar)

        container = QWidget()
        container.setLayout(QHBoxLayout())

        buttons = QDialogButtonBox(
            orientation=Qt.Horizontal,
            standardButtons=QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
        )

        empty = QWidget()
        empty.setFixedWidth(1)

        container.layout().addWidget(buttons)
        container.layout().addWidget(empty)

        buttons.accepted.connect(self.__accepted)
        buttons.rejected.connect(self.__rejected)

        empty = QWidget()
        empty.setFixedHeight(1)

        self.layout().addWidget(container)
        self.layout().addWidget(empty)

        self._executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        if InternalLibrariesManagerDialog._packages is None:
            self._f_pypi_addons = self._executor.submit(list_available_versions)
        else:
            self._f_pypi_addons = concurrent.futures.Future()
            self._f_pypi_addons.set_result(InternalLibrariesManagerDialog._packages)

        self._f_pypi_addons.add_done_callback(
            method_queued(self._set_packages, (object,))
        )

        self.__progress = None  # type: Optional[QProgressDialog]
        self.__thread = None
        self.__installer = None

        if not self._f_pypi_addons.done():
            self.__progressDialog()

        self.__is_app_to_be_closed = False

    def set_is_app_to_be_closed(self, is_app_to_be_closed=True):
        self.__is_app_to_be_closed = is_app_to_be_closed

    @Slot(str, str)
    def __show_error_for_query(self, text, error_details):
        message_error(text, title="Error", details=error_details)

    @Slot(object)
    def add_package(self, installable):
        # type: (Installable) -> None
        if installable.name in {p.name for p in self._packages}:
            return
        else:
            packages = self._packages + [installable]
        self.set_packages(packages)

    def __progressDialog(self):
        if self.__progress is None:
            self.__progress = QProgressDialog(
                self,
                minimum=0, maximum=0,
                labelText=self.tr("Retrieving package list"),
                sizeGripEnabled=False,
                windowTitle="Progress",
            )
            self.__progress.setWindowModality(Qt.WindowModal)
            self.__progress.canceled.connect(self.reject)
            self.__progress.hide()

        return self.__progress

    @Slot(object)
    def _set_packages(self, f):
        if self.__progress is not None:
            self.__progress.hide()
            self.__progress.deleteLater()
            self.__progress = None

        try:
            packages = f.result()
        except Exception as err:
            message_warning(
                "Could not retrieve package list",
                title="Error",
                informative_text=str(err),
                parent=self
            )
            log.error(str(err), exc_info=True)
            packages = []
        else:
            InternalLibrariesManagerDialog._packages = packages

        self.set_packages(packages)

    @Slot(object)
    def set_packages(self, installable):
        # type: (List[Installable]) -> None
        self._packages = packages = installable  # type: List[Installable]
        installed = list_installed_internal_libraries()
        dists = {dist.project_name: dist for dist in installed}
        packages = {pkg.name: pkg for pkg in packages}

        # For every pypi available distribution not listed by
        # list_installed_addons, check if it is actually already
        # installed.
        ws = pkg_resources.WorkingSet()
        for pkg_name in set(packages.keys()).difference(set(dists.keys())):
            try:
                d = ws.find(pkg_resources.Requirement.parse(pkg_name))
            except pkg_resources.VersionConflict:
                pass
            except ValueError:
                # Requirements.parse error ?
                pass
            else:
                if d is not None:
                    dists[d.project_name] = d

        project_names = unique(
            itertools.chain(packages.keys(), dists.keys())
        )

        items = []
        for name in project_names:
            if name in dists and name in packages:
                item = Installed(packages[name], dists[name])
            elif name in dists:
                item = Installed(None, dists[name])
            elif name in packages:
                item = Available(packages[name])
            else:
                assert False
            items.append(item)

        self.internallibrarieswidget.set_items(items)

    def showEvent(self, event):
        super().showEvent(event)

        if not self._f_pypi_addons.done() and self.__progress is not None:
            QTimer.singleShot(0, self.__progress.show)

    def done(self, retcode):
        super().done(retcode)
        self._f_pypi_addons.cancel()
        self._executor.shutdown(wait=False)
        if self.__thread is not None:
            self.__thread.quit()
            self.__thread.wait(1000)

    def closeEvent(self, event):
        super().closeEvent(event)
        if self.__progress is not None:
            self.__progress.hide()
        self._f_pypi_addons.cancel()
        self._executor.shutdown(wait=False)

        if self.__thread is not None:
            self.__thread.quit()
            self.__thread.wait(1000)

    ADDON_EXTENSIONS = ('.zip', '.whl', '.tar.gz')

    def dragEnterEvent(self, event):
        urls = event.mimeData().urls()
        if any((OSX_NSURL_toLocalFile(url) or url.toLocalFile())
               .endswith(self.ADDON_EXTENSIONS) for url in urls):
            event.acceptProposedAction()

    def dropEvent(self, event):
        """Allow dropping add-ons (zip or wheel archives) on this dialog to
        install them"""
        packages = []
        names = []
        for url in event.mimeData().urls():
            path = OSX_NSURL_toLocalFile(url) or url.toLocalFile()
            if path.endswith(self.ADDON_EXTENSIONS):
                name, vers, summary, descr = (get_meta_from_archive(path) or
                                              (os.path.basename(path), '', '', ''))
                names.append(cleanup(name))
                packages.append(
                    Installable(name, vers, summary,
                                descr or summary, path, [path]))
        future = concurrent.futures.Future()
        future.set_result((InternalLibrariesManagerDialog._packages or []) + packages)
        self._set_packages(future)
        self.internallibrarieswidget.set_install_projects(names)

    def __rejected(self):
        self.reject()

    def __accepted(self):
        steps = self.internallibrarieswidget.item_state()

        if steps:
            # Move all uninstall steps to the front
            steps = sorted(
                steps, key=lambda step: 0 if step[0] == Uninstall else 1
            )
            self.__installer = Installer(steps=steps)
            self.__thread = QThread(self)
            self.__thread.start()

            self.__installer.moveToThread(self.__thread)
            self.__installer.finished.connect(self.__on_installer_finished)
            self.__installer.error.connect(self.__on_installer_error)

            progress = self.__progressDialog()
            self.__installer.installStatusChanged.connect(progress.setLabelText)
            progress.show()
            progress.setLabelText("Installing")

            self.__installer.start()
        else:
            self.accept()

    def __on_installer_error(self, command, pkg, retcode, output):
        message_error(
            "An error occurred while running a subprocess", title="Error",
            informative_text="{} exited with non zero status.".format(command),
            details="".join(output),
            parent=self
        )
        self.reject()

    def __on_installer_finished(self):
        if self.__is_app_to_be_closed:
            message = "Click Ok to restart OASYS for changes to take effect."
            message_information(message, parent=self)
            self.accept()
            sys.exit(0)
        else:
            self.accept()

import platform

def list_available_internal_libraries():
    if is_auto_update:

        system, node, release, version, machine, processor = platform.uname()

        packages = []
        for library in internal_libraries_list:
            try:
                info = library["info"]

                if not ("Debian 3" in version and info["name"]=="PyQt5"):
                    packages.append(
                    Installable(info["name"], info["version"],
                                info["summary"], info["description"],
                                info["package_url"],
                                info["package_url"])
                )
            except (TypeError, KeyError):
                continue  # skip invalid packages

        return packages

def list_available_versions():
    """
    List add-ons available.
    """

    if is_auto_update:
        internal_libraries = internal_libraries_list
    else:
        internal_libraries = []

    # query pypi.org for installed add-ons that are not in our list
    installed = list_installed_internal_libraries()
    missing = set(dist.project_name for dist in installed) - \
              set(a.get("info", {}).get("name", "") for a in internal_libraries)
    for p in missing:
        response = requests.get(PYPI_API_JSON.format(name=p))
        if response.status_code != 200:
            continue
        internal_libraries.append(response.json())

    packages = []
    for internal_library in internal_libraries:
        try:
            info = internal_library["info"]
            packages.append(
                Installable(info["name"], info["version"],
                            info["summary"], info["description"],
                            info["package_url"],
                            info["package_url"])
            )
        except (TypeError, KeyError):
            continue  # skip invalid packages

    return packages

def list_installed_internal_libraries():
    workingset = pkg_resources.WorkingSet(sys.path)

    installed_internal_libraries = []

    for internal_library in INTERNAL_LIBRARIES:
        try:
            installed_internal_libraries.append(workingset.by_key[internal_library.lower()])
        except KeyError:
            pass

    return installed_internal_libraries




# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.2.140'
version = '1.2.140'
full_version = '1.2.140'
git_revision = ''
release = True

if not release:
    version = full_version
    short_version += ".dev"



class AlreadyInitializedError(ValueError):
    def __init__(self, message=None): super(AlreadyInitializedError, self).__init__(message)

class GenericRegistry(object):
    _NO_APPLICATION = "<NO APPLICATION>"

    def __init__(self, registry_name):
        self.__registry_name = registry_name
        self.__registry = {self._NO_APPLICATION: None}

    def register_instance(self, instance, application_name=None, replace=False):
        if instance is None: raise ValueError(self.__registry_name + " Instance is None")

        application_name = self.__get_application_name(application_name)

        if application_name in self.__registry.keys():
            if self.__registry[application_name] is None or replace==True: self.__registry[application_name] = instance
            else: raise AlreadyInitializedError(self.__registry_name + " Instance already initialized")
        else: self.__registry[application_name] = instance

    def reset(self, application_name=None):
        application_name = self.__get_application_name(application_name)

        if application_name in self.__registry.keys(): self.__registry[self.__get_application_name(application_name)] = None
        else: raise ValueError(self.__registry_name + " Instance not existing")

    def get_instance(self, application_name=None):
        application_name = self.__get_application_name(application_name)

        if application_name in self.__registry.keys(): return self.__registry[self.__get_application_name(application_name)]
        else: raise ValueError(self.__registry_name + " Instance not existing")

    def __get_application_name(self, application_name):
        return self._NO_APPLICATION if application_name is None else application_name
